
"""
 * ApMon - Application Monitoring Tool
 * Version: 2.2.16
 *
 * Copyright (C) 2006 California Institute of Technology
 *
 * Permission is hereby granted, free of charge, to use, copy and modify 
 * this software and its documentation (the "Software") for any
 * purpose, provided that existing copyright notices are retained in 
 * all copies and that this notice is included verbatim in any distributions
 * or substantial portions of the Software. 
 * This software is a part of the MonALISA framework (http://monalisa.cacr.caltech.edu).
 * Users of the Software are asked to feed back problems, benefits,
 * and/or suggestions about the software to the MonALISA Development Team
 * (developers@monalisa.cern.ch). Support for this software - fixing of bugs,
 * incorporation of new features - is done on a best effort basis. All bug
 * fixes and enhancements will be made available under the same terms and
 * conditions as the original software,

 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY DERIVATIVES THEREOF,
 * EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT. THIS SOFTWARE IS
 * PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE NO
 * OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS.
"""
from __future__ import print_function


import os
import re
import string
import socket
from Logger import *

"""
Class ProcInfo
extracts information from the proc/ filesystem for system and job monitoring
"""


class ProcInfo:

    # ProcInfo constructor
    def __init__(self, logger):
        self.DATA = {}             # monitored data that is going to be reported
        self.LAST_UPDATE_TIME = 0  # when the last measurement was done
        self.JOBS = {}             # jobs that will be monitored
        self.logger = logger	    # use the given logger
        self.readGenericInfo()

    # This should be called from time to time to update the monitored data,
    # but not more often than once a second because of the resolution of time()
    def update(self):
        if self.LAST_UPDATE_TIME == int(time.time()):
            self.logger.log(Logger.NOTICE, "ProcInfo: update() called too often!")
            return
        self.readStat()
        self.readMemInfo()
        self.readUptimeAndLoadAvg()
        self.countProcesses()
        self.readNetworkInfo()
        self.readNetStat()
        for pid in self.JOBS.keys():
            self.readJobInfo(pid)
            self.readJobDiskUsage(pid)
        self.LAST_UPDATE_TIME = int(time.time())
        self.DATA['TIME'] = int(time.time())

    # Call this to add another PID to be monitored
    def addJobToMonitor(self, pid, workDir):
        self.JOBS[pid] = {}
        self.JOBS[pid]['WORKDIR'] = workDir
        self.JOBS[pid]['DATA'] = {}
        # print self.JOBS

    # Call this to stop monitoring a PID
    def removeJobToMonitor(self, pid):
        if pid in self.JOBS:
            del self.JOBS[pid]

    # Return a filtered hash containting the system-related parameters and values
    def getSystemData(self, params, prevDataRef):
        return self.getFilteredData(self.DATA, params, prevDataRef)

    # Return a filtered hash containing the job-related parameters and values
    def getJobData(self, pid, params):
        if pid not in self.JOBS:
            return []
        return self.getFilteredData(self.JOBS[pid]['DATA'], params)

    ############################################################################################
    # internal functions for system monitoring
    ############################################################################################

    # this has to be run twice (with the $lastUpdateTime updated) to get some useful results
    # the information about blocks_in/out and swap_in/out isn't available for 2.6 kernels (yet)
    def readStat(self):
        try:
            FSTAT = open('/proc/stat')
            line = FSTAT.readline()
            while line != '':
                if line.startswith("cpu "):
                    elem = re.split("\s+", line)
                    self.DATA['raw_cpu_usr'] = float(elem[1])
                    self.DATA['raw_cpu_nice'] = float(elem[2])
                    self.DATA['raw_cpu_sys'] = float(elem[3])
                    self.DATA['raw_cpu_idle'] = float(elem[4])
                    self.DATA['raw_cpu_iowait'] = float(elem[5])
                elif line.startswith("page"):
                    elem = line.split()
                    self.DATA['raw_blocks_in'] = float(elem[1])
                    self.DATA['raw_blocks_out'] = float(elem[2])
                elif line.startswith('swap'):
                    elem = line.split()
                    self.DATA['raw_swap_in'] = float(elem[1])
                    self.DATA['raw_swap_out'] = float(elem[2])
                elif line.startswith('intr'):
                    elem = line.split()
                    self.DATA['raw_interrupts'] = float(elem[1])
                elif line.startswith('ctxt'):
                    elem = line.split()
                    self.DATA['raw_context_switches'] = float(elem[1])
                line = FSTAT.readline()
            FSTAT.close()
        except IOError as ex:
            self.logger.log(Logger.ERROR, "ProcInfo: cannot open /proc/stat")
            return
        try:
            # blocks_in/out and swap_in/out are moved to /proc/vmstat in >2.5 kernels
            FVMSTAT = open('/proc/vmstat')
            line = FVMSTAT.readline()
            while line != '':
                elem = re.split("\s+", line)
                if line.startswith("pgpgin "):
                    self.DATA['raw_blocks_in'] = float(elem[1])
                elif line.startswith("pgpgout "):
                    self.DATA['raw_blocks_out'] = float(elem[1])
                elif line.startswith("pswpin "):
                    self.DATA['raw_swap_in'] = float(elem[1])
                elif line.startswith("pswpout "):
                    self.DATA['raw_swap_out'] = float(elem[1])
                line = FVMSTAT.readline()
            FVMSTAT.close()
        except IOError as ex:
            self.logger.log(Logger.NOTICE, "ProcInfo: cannot open /proc/vmstat")

    # sizes are reported in MB (except _usage that is in percent).
    def readMemInfo(self):
        try:
            FMEM = open('/proc/meminfo')
            line = FMEM.readline()
            while line != '':
                elem = re.split("\s+", line)
                if line.startswith("MemFree:"):
                    self.DATA['mem_free'] = float(elem[1]) / 1024.0
                if line.startswith("MemTotal:"):
                    self.DATA['total_mem'] = float(elem[1]) / 1024.0
                if line.startswith("SwapFree:"):
                    self.DATA['swap_free'] = float(elem[1]) / 1024.0
                if line.startswith("SwapTotal:"):
                    self.DATA['total_swap'] = float(elem[1]) / 1024.0
                if line.startswith("Buffers:"):
                    self.DATA['mem_buffers'] = float(elem[1]) / 1024.0
                if line.startswith("Cached:"):
                    self.DATA['mem_cached'] = float(elem[1]) / 1024.0
                line = FMEM.readline()
            FMEM.close()
            if 'mem_free' in self.DATA and 'mem_buffers' in self.DATA and 'mem_cached' in self.DATA:
                self.DATA['mem_actualfree'] = self.DATA['mem_free'] + self.DATA['mem_buffers'] + self.DATA['mem_cached']
            if 'total_mem' in self.DATA and 'mem_actualfree' in self.DATA:
                self.DATA['mem_used'] = self.DATA['total_mem'] - self.DATA['mem_actualfree']
            if 'total_swap' in self.DATA and 'swap_free' in self.DATA:
                self.DATA['swap_used'] = self.DATA['total_swap'] - self.DATA['swap_free']
            if 'mem_used' in self.DATA and 'total_mem' in self.DATA and self.DATA['total_mem'] > 0:
                self.DATA['mem_usage'] = 100.0 * self.DATA['mem_used'] / self.DATA['total_mem']
            if 'swap_used' in self.DATA and 'total_swap' in self.DATA and self.DATA['total_swap'] > 0:
                self.DATA['swap_usage'] = 100.0 * self.DATA['swap_used'] / self.DATA['total_swap']
        except IOError as ex:
            self.logger.log(Logger.ERROR, "ProcInfo: cannot open /proc/meminfo")
            return

    # read system load average
    def readLoadAvg(self):
        try:
            FAVG = open('/proc/loadavg')
            line = FAVG.readline()
            FAVG.close()
            elem = re.split("\s+", line)
            self.DATA['load1'] = float(elem[0])
            self.DATA['load5'] = float(elem[1])
            self.DATA['load15'] = float(elem[2])
        except IOError as ex:
            self.logger.log(Logger.ERROR, "ProcInfo: cannot open /proc/meminfo")
            return

    # read system load average on Darwin
    def darwin_readLoadAvg(self):
        try:
            LOAD_AVG = os.popen('sysctl vm.loadavg')
            line = LOAD_AVG.readline()
            LOAD_AVG.close()
            elem = re.split("\s+", line)
            self.DATA['load1'] = float(elem[1])
            self.DATA['load5'] = float(elem[2])
            self.DATA['load15'] = float(elem[3])
        except IOError as ex:
            self.logger.log(Logger.ERROR, "ProcInfo: cannot run 'sysctl vm.loadavg")
            return

    # read the number of processes currently running on the system
    def countProcesses(self):
        """
        # old version
        nr = 0
        try:
            for file in os.listdir("/proc"):
                if re.match("\d+", file):
                    nr += 1
            self.DATA['processes'] = nr
        except IOError, ex:
            self.logger.log(Logger.ERROR, "ProcInfo: cannot open /proc to count processes")
            return
        """
        # new version
        total = 0
        states = {'D':0, 'R':0, 'S':0, 'T':0, 'Z':0}
        try:
            output = os.popen('ps -A -o state')
            line = output.readline()
            while line != '':
                if line[0] in states:
                    states[line[0]] = states[line[0]] + 1
                else:
                    states[line[0]] = 1
                total = total + 1
                line = output.readline()
            output.close()
            self.DATA['processes'] = total
            for key in states.keys():
                self.DATA['processes_'+key] = states[key]
        except IOError as ex:
            self.logger.log(Logger.ERROR, "ProcInfo: cannot get output from ps command")
            return

    # reads the IP, hostname, cpu_MHz, uptime
    def readGenericInfo(self):
        self.DATA['hostname'] = socket.getfqdn()
        try:
            output = os.popen('/sbin/ifconfig -a')
            eth, ip = '', ''
            line = output.readline()
            while line != '':
                line = line.strip()
                if line.startswith("eth"):
                    elem = line.split()
                    eth = elem[0]
                    ip = ''
                if len(eth) > 0 and line.startswith("inet addr:"):
                    ip = re.match("inet addr:(\d+\.\d+\.\d+\.\d+)", line).group(1)
                    self.DATA[eth + '_ip'] = ip
                    eth = ''
                line = output.readline()
            output.close()
        except IOError as ex:
            self.logger.log(Logger.ERROR, "ProcInfo: cannot get output from /sbin/ifconfig -a")

        try:
            no_cpus = 0
            FCPU = open('/proc/cpuinfo')
            line = FCPU.readline()
            while line != '':
                if line.startswith("cpu MHz"):
                    self.DATA['cpu_MHz'] = float(re.match("cpu MHz\s+:\s+(\d+\.?\d*)", line).group(1))
                    no_cpus += 1

                if line.startswith("vendor_id"):
                    self.DATA['cpu_vendor_id'] = re.match("vendor_id\s+:\s+(.+)", line).group(1)

                if line.startswith("cpu family"):
                    self.DATA['cpu_family'] = re.match("cpu family\s+:\s+(.+)", line).group(1)

                if line.startswith("model") and not line.startswith("model name") :
                    self.DATA['cpu_model'] = re.match("model\s+:\s+(.+)", line).group(1)

                if line.startswith("model name"):
                    self.DATA['cpu_model_name'] = re.match("model name\s+:\s+(.+)", line).group(1)

                if line.startswith("bogomips"):
                    self.DATA['bogomips'] = float(re.match("bogomips\s+:\s+(\d+\.?\d*)", line).group(1))

                line = FCPU.readline()
            FCPU.close()
            self.DATA['no_CPUs'] = no_cpus
        except IOError as ex:
            self.logger.log(Logger.ERROR, "ProcInfo: cannot open /proc/cpuinfo")

        # try to determine the kernel version
        try:
            output = os.popen('uname -r')
            line = output.readline().replace("\n", "")
            self.DATA['kernel_version'] = line
            output.close()
        except IOError as ex:
            self.logger.log(Logger.ERROR, "ProcInfo: cannot get kernel version with 'uname -r'")

        # try to determine the platform
        try:
            output = os.popen('uname -m 2>/dev/null || uname')
            line = output.readline().replace("\n", "")
            self.DATA['platform'] = line
            output.close()
        except IOError as ex:
            self.logger.log(Logger.ERROR, "ProcInfo: cannot get platform with 'uname -m'")

        # try to determine the OS type
        osType = None
        try:
            output = os.popen('env PATH=$PATH:/bin:/usr/bin lsb_release -d 2>/dev/null')
            line = output.readline().replace("\n", "")
            mo = re.match("Description:\s*(.*)", line)
            if not mo is None:
                osType = mo.group(1)
            output.close()
        except IOError as ex:
            pass
        # if lsb_release didn't work, try again...
        if osType is None:
            for file in ["/etc/redhat-release", "/etc/debian_version", "/etc/SuSE-release", "/etc/slackware-version", "/etc/gentoo-release", "/etc/mandrake-release", "/etc/mandriva-release", "/etc/issue"]:
                try:
                    REL = open(file)
                    line = REL.readline().replace("\n", "")
                    if len(line) > 0:
                        osType = line
                    REL.close()
                    break
                except IOError as ex:
                    pass
        # if none of these, try just uname -s
        if osType is None:
            try:
                output = os.popen('uname -s')
                line = output.readline().replace("\n", "")
                osType = line
                output.close()
            except IOError as ex:
                pass
        if not osType is None:
            self.DATA['os_type'] = osType
        else:
            self.logger.log(Logger.ERROR, "ProcInfo: cannot determine operating system type")

    # read system's uptime and load average. Time is reported as a floating number, in days.
    # It uses the 'uptime' command which's output looks like these:
    # 19:55:37 up 11 days, 18:57,  1 user,  load average: 0.00, 0.00, 0.00
    # 18:42:31 up 87 days, 18:10,  9 users,  load average: 0.64, 0.84, 0.80
    # 6:42pm  up 7 days  3:08,  7 users,  load average: 0.18, 0.14, 0.10
    # 6:42pm  up 33 day(s),  1:54,  1 user,  load average: 0.01, 0.00, 0.00
    # 18:42  up 7 days,  3:45, 2 users, load averages: 1.10 1.11 1.06
    # 18:47:41  up 7 days,  4:35, 19 users,  load average: 0.66, 0.44, 0.41
    # 11:57am  up   2:21,  22 users,  load average: 0.59, 0.93, 0.73
    def readUptimeAndLoadAvg(self):
        try:
            output = os.popen('uptime')
            line = output.readline().replace("\n", "")
            mo = re.match(".*up\s+((\d+)\s+day[ (s),]+)?(\d+)(:(\d+))?[^\d]+(\d+)[^\d]+([\d\.]+)[^\d]+([\d\.]+)[^\d]+([\d\.]+)", line)
            if mo is None:
                self.logger.log(Logger.ERROR, "ProcInfo: got nonparsable output from uptime: "+line)
                return
            (days, hour, min, users, load1, load5, load15) = (mo.group(2), mo.group(3), mo.group(5), float(mo.group(6)), mo.group(7), mo.group(8), mo.group(9))
            if days is None:
                days = 0.0
            if min is None:
                (min, hour) = (hour, 0.0)
            uptime = float(days) + float(hour) / 24.0 + float(min) / 1440.0
            self.DATA['uptime'] = float(uptime)
            self.DATA['logged_users'] = float(users) # this is currently not reported
            self.DATA['load1'] = float(load1)
            self.DATA['load5'] = float(load5)
            self.DATA['load15'] = float(load15)
            output.close()
        except IOError as ex:
            self.logger.log(Logger.ERROR, "ProcInfo: cannot get output from uptime command")

    # do a difference with overflow check and repair
    # the counter is unsigned 32 or 64   bit
    @staticmethod
    def diffWithOverflowCheck(self, new, old):
        if new >= old:
            return new - old
        else:
            max = (1 << 31) * 2  # 32 bits
            if old >= max:
                max = (1 << 63) * 2  # 64 bits
            return new - old + max

    # read network information like transfered kBps and nr. of errors on each interface
    def readNetworkInfo(self):
        try:
            FNET = open('/proc/net/dev')
            line = FNET.readline()
            while line != '':
                m = re.match("\s*eth(\d):(\d+)\s+\d+\s+(\d+)\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+\s+(\d+)\s+\d+\s+(\d+)", line)
                if m != None:
                    self.DATA['raw_eth'+m.group(1)+'_in'] = float(m.group(2))
                    self.DATA['raw_eth'+m.group(1)+'_out'] = float(m.group(4))
                    self.DATA['raw_eth'+m.group(1)+'_errs'] = int(m.group(3)) + int(m.group(5))
                line = FNET.readline()
            FNET.close()
        except IOError as ex:
            self.logger.log(Logger.ERROR, "ProcInfo: cannot open /proc/net/dev")
            return

    # run nestat and collect sockets info (tcp, udp, unix) and connection states for tcp sockets from netstat
    def readNetStat(self):
        try:
            output = os.popen('netstat -an 2>/dev/null')
            sockets = { 'sockets_tcp': 0, 'sockets_udp': 0, 'sockets_unix': 0, 'sockets_icm': 0}
            tcp_details = {'sockets_tcp_ESTABLISHED': 0, 'sockets_tcp_SYN_SENT': 0,
                           'sockets_tcp_SYN_RECV': 0, 'sockets_tcp_FIN_WAIT1': 0, 'sockets_tcp_FIN_WAIT2': 0,
                           'sockets_tcp_TIME_WAIT': 0, 'sockets_tcp_CLOSED': 0, 'sockets_tcp_CLOSE_WAIT': 0,
                           'sockets_tcp_LAST_ACK': 0, 'sockets_tcp_LISTEN': 0, 'sockets_tcp_CLOSING': 0,
                           'sockets_tcp_UNKNOWN': 0}
            line = output.readline()
            while line != '':
                arg = string.split(line)
                proto = arg[0]
                if proto.find('tcp') == 0:
                    sockets['sockets_tcp'] += 1
                    state = arg[len(arg)-1]
                    key = 'sockets_tcp_'+state
                    if key in tcp_details:
                        tcp_details[key] += 1
                if proto.find('udp') == 0:
                    sockets['sockets_udp'] += 1
                if proto.find('unix') == 0:
                    sockets['sockets_unix'] += 1
                if proto.find('icm') == 0:
                    sockets['sockets_icm'] += 1

                line = output.readline()
            output.close()

            for key in sockets.keys():
                    self.DATA[key] = sockets[key]
            for key in tcp_details.keys():
                    self.DATA[key] = tcp_details[key]
        except IOError as ex:
                self.logger.log(Logger.ERROR, "ProcInfo: cannot get output from netstat command")
                return

    ##############################################################################################
    # job monitoring related functions
    ##############################################################################################

    # internal function that gets the full list of children (pids) for a process (pid)
    def getChildren(self, parent):
        pidmap = {}
        try:
            output = os.popen('ps -A -o "pid ppid"')
            line = output.readline() # skip headers
            line = output.readline()
            while(line != ''):
                line = line.strip()
                elem = re.split("\s+", line)
                pidmap[int(elem[0])] = int(elem[1])
                line = output.readline()
            output.close()
        except IOError as ex:
            self.logger.log(Logger.ERROR, "ProcInfo: cannot execute ps -A -o \"pid ppid\"")

        if parent not in pidmap:
            self.logger.log(Logger.INFO, 'ProcInfo: No job with pid='+str(parent))
            self.removeJobToMonitor(parent)
            return []

        children = [parent]
        i = 0
        while i < len(children):
            prnt = children[i]
            for (pid, ppid) in pidmap.items():
                if ppid == prnt:
                    children.append(pid)
                i += 1
        return children

    # internal function that parses a time formatted like "days-hours:min:sec" and returns the corresponding
    # number of seconds.
    def parsePSTime(self, my_time):
        my_time = my_time.strip()
        m = re.match("(\d+)-(\d+):(\d+):(\d+)", my_time)
        if m != None:
            return int(m.group(1)) * 24 * 3600 + int(m.group(2)) * 3600 + int(m.group(3)) * 60 + int(m.group(4))
        else:
            m = re.match("(\d+):(\d+):(\d+)", my_time)
            if m != None:
                return int(m.group(1)) * 3600 + int(m.group(2)) * 60 + int(m.group(3))
            else:
                m = re.match("(\d+):(\d+)", my_time)
                if m != None:
                    return int(m.group(1)) * 60 + int(m.group(2))
                else:
                    return None

    # read information about this the JOB_PID process
    # memory sizes are given in KB
    def readJobInfo(self, pid):
        if (pid == '') or pid not in self.JOBS:
            return
        children = self.getChildren(pid)
        if len(children) == 0:
            self.logger.log(Logger.INFO, "ProcInfo: Job with pid="+str(pid)+" terminated; removing it from monitored jobs.")
            # print ":("
            self.removeJobToMonitor(pid)
            return
        try:
            JSTATUS = os.popen("ps --no-headers --pid " + ",".join([repr(child) for child in  children]) + " -o pid,etime,time,%cpu,%mem,rsz,vsz,comm")
            mem_cmd_map = {}
            etime, cputime, pcpu, pmem, rsz, vsz, comm, fd = None, None, None, None, None, None, None, None
            line = JSTATUS.readline()
            while line != '':
                line = line.strip()
                m = re.match("(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(.+)", line)
                if m != None:
                    apid, etime1, cputime1, pcpu1, pmem1, rsz1, vsz1, comm1 = m.group(1), m.group(2), m.group(3), m.group(4), m.group(5), m.group(6), m.group(7), m.group(8)
                    sec = self.parsePSTime(etime1)
                    if (not sec is None) and (sec > self.addIfValid(etime, 0)): 	# the elapsed time is the maximum of all elapsed
                        etime = sec
                    sec = self.parsePSTime(cputime1)		# times corespornding to all child processes.
                    cputime = self.addIfValid(cputime, sec)	# total cputime is the sum of cputimes for all processes.
                    pcpu = self.addIfValid(pcpu, float(pcpu1))	# total %cpu is the sum of all children %cpu.
                    if repr(pmem1)+" "+repr(rsz1)+" "+repr(vsz1)+" "+repr(comm1) not in mem_cmd_map:
                        # it's the first thread/process with this memory footprint; add it.
                        mem_cmd_map[repr(pmem1)+" "+repr(rsz1)+" "+repr(vsz1)+" "+repr(comm1)] = 1
                        pmem = self.addIfValid(pmem, float(pmem1))
                        rsz = self.addIfValid(rsz, int(rsz1))
                        vsz = self.addIfValid(vsz, int(vsz1))
                        fd = self.addIfValid(fd, self.countOpenFD(apid))
                    # else not adding memory usage
                line = JSTATUS.readline()
            JSTATUS.close()
            if not etime is None:
                self.JOBS[pid]['DATA']['run_time'] = etime
            if not cputime is None:
                self.JOBS[pid]['DATA']['cpu_time'] = cputime
            if not pcpu is None:
                self.JOBS[pid]['DATA']['cpu_usage'] = pcpu
            if not pmem is None:
                self.JOBS[pid]['DATA']['mem_usage'] = pmem
            if not rsz is None:
                self.JOBS[pid]['DATA']['rss'] = rsz
            if not vsz is None:
                self.JOBS[pid]['DATA']['virtualmem'] = vsz
            if not fd is None:
                self.JOBS[pid]['DATA']['open_files'] = fd
        except IOError as ex:
            self.logger.log(Logger.ERROR, "ProcInfo: cannot execute ps --no-headers -eo \"pid ppid\"")

    # return the sum of the two given parameters (or None if the case)
    def addIfValid(self, sum, toAdd):
        if toAdd is None:
            return sum
        if sum is None:
            return toAdd
        return sum + toAdd

    # count the number of open files for the given pid
    def countOpenFD(self, pid):
        dir = '/proc/'+str(pid)+'/fd'
        if os.access(dir, os.F_OK):
            if os.access(dir, os.X_OK):
                list = os.listdir(dir)
                open_files = len(list)
                if pid == os.getpid():
                    open_files -= 2
                self.logger.log(Logger.DEBUG, "Counting open_files for "+ repr(pid) +": "+ str(len(list)) +" => " + repr(open_files) + " open_files")
                return open_files
            else:
                self.logger.log(Logger.ERROR, "ProcInfo: cannot count the number of opened files for job "+repr(pid))
        else:
            self.logger.log(Logger.ERROR, "ProcInfo: job "+repr(pid)+" dosen't exist")
        # failed
        return None

    # if there is an work directory defined, then compute the used space in that directory
    # and the free disk space on the partition to which that directory belongs
    # sizes are given in MB
    def readJobDiskUsage(self, pid):
        if (pid == '') or pid not in self.JOBS:
            return
        workDir = self.JOBS[pid]['WORKDIR']
        if workDir == '':
            return
        try:
            DU = os.popen("du -Lsck " + workDir + " | tail -1 | cut -f 1")
            line = DU.readline()
            self.JOBS[pid]['DATA']['workdir_size'] = int(line) / 1024.0
        except IOError as ex:
            self.logger.log(Logger.ERROR, "ERROR", "ProcInfo: cannot run du to get job's disk usage for job "+repr(pid))
        try:
            DF = os.popen("df -k "+workDir+" | tail -1")
            line = DF.readline().strip()
            m = re.match("\S+\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)%", line)
            if m != None:
                self.JOBS[pid]['DATA']['disk_total'] = float(m.group(1)) / 1024.0
                self.JOBS[pid]['DATA']['disk_used']  = float(m.group(2)) / 1024.0
                self.JOBS[pid]['DATA']['disk_free']  = float(m.group(3)) / 1024.0
                self.JOBS[pid]['DATA']['disk_usage'] = float(m.group(4)) / 1024.0
            DF.close()
        except IOError as ex:
            self.logger.log(Logger.ERROR, "ERROR", "ProcInfo: cannot run df to get job's disk usage for job "+repr(pid))

    # create cummulative parameters based on raw params like cpu_, blocks_, swap_, or ethX_
    def computeCummulativeParams(self, dataRef, prevDataRef):
        if prevDataRef == {}:
            for key in dataRef.keys():
                if key.find('raw_') == 0:
                    prevDataRef[key] = dataRef[key]
            prevDataRef['TIME'] = dataRef['TIME']
            return

        # cpu -related params
        if ('raw_cpu_usr' in dataRef) and ('raw_cpu_usr' in prevDataRef):
            diff={}
            cpu_sum = 0
            for param in ['cpu_usr', 'cpu_nice', 'cpu_sys', 'cpu_idle', 'cpu_iowait']:
                diff[param] = self.diffWithOverflowCheck(dataRef['raw_'+param], prevDataRef['raw_'+param])
                cpu_sum += diff[param]
            for param in ['cpu_usr', 'cpu_nice', 'cpu_sys', 'cpu_idle', 'cpu_iowait']:
                if cpu_sum != 0:
                    dataRef[param] = 100.0 * diff[param] / cpu_sum
                else:
                    del dataRef[param]
            if cpu_sum != 0:
                dataRef['cpu_usage'] = 100.0 * (cpu_sum - diff['cpu_idle']) / cpu_sum
            else:
                del dataRef['cpu_usage']
            # add the other parameters
            for param in ['interrupts', 'context_switches']:
                if('raw_'+param in prevDataRef and 'raw_'+param in dataRef):
                    dataRef[param] = self.diffWithOverflowCheck(dataRef['raw_'+param], prevDataRef['raw_'+param])
                else:
                    del dataRef[param]

        # swap, blocks, context switches, interrupts
        interval = dataRef['TIME'] - prevDataRef['TIME']
        for param in ['blocks_in', 'blocks_out', 'swap_in', 'swap_out', 'interrupts', 'context_switches']:
            if((interval != 0) and 'raw_'+param in prevDataRef and 'raw_'+param in dataRef):
                diff = self.diffWithOverflowCheck(dataRef['raw_'+param], prevDataRef['raw_'+param])
                dataRef[param+'_R'] = diff / interval
            else:
                del dataRef[param+'_R']

        # eth - related params
        for rawParam in dataRef.keys():
            if (rawParam.find('raw_eth') == 0) and rawParam in prevDataRef:
                param = rawParam.split('raw_')[1]
                if interval != 0:
                    dataRef[param] = self.diffWithOverflowCheck(dataRef[rawParam], prevDataRef[rawParam]) # absolute difference
                    if param.find('_errs') == -1:
                        dataRef[param] = dataRef[param] / interval / 1024.0 # if it's _in or _out, compute in KB/sec
                else:
                    del dataRef[param]

        # copy contents of the current data values to the
        for param in dataRef.keys():
            if param.find('raw_') == 0:
                prevDataRef[param] = dataRef[param]
        prevDataRef['TIME'] = dataRef['TIME']

    # Return a hash containing (param,value) pairs with existing values from the requested ones
    def getFilteredData(self, dataHash, paramsList, prevDataHash = None):

        if not prevDataHash is None:
            self.computeCummulativeParams(dataHash, prevDataHash)

        result = {}
        for param in paramsList:
            if param == 'net_sockets':
                for key in dataHash.keys():
                    if key.find('sockets') == 0 and key.find('sockets_tcp_') == -1:
                        result[key] = dataHash[key]
            elif param == 'net_tcp_details':
                for key in dataHash.keys():
                    if key.find('sockets_tcp_') == 0:
                        result[key] = dataHash[key]

            m = re.match("^net_(.*)$", param)
            if m == None:
                m = re.match("^(ip)$", param)
            if m != None:
                net_param = m.group(1)
                # self.logger.log(Logger.DEBUG, "Querying param "+net_param)
                for key, value in dataHash.items():
                    m = re.match("eth\d_"+net_param, key)
                    if m != None:
                        result[key] = value
            else:
                if param == 'processes':
                    for key in dataHash.keys():
                        if key.find('processes') == 0:
                            result[key] = dataHash[key]
                elif param in dataHash:
                    result[param] = dataHash[param]
        sorted_result = []
        keys = result.keys()
        keys.sort()
        for key in keys:
            sorted_result.append((key, result[key]))
        return sorted_result

######################################################################################
# self test

if __name__ == '__main__':
    logger = Logger.Logger(Logger.DEBUG)
    pi = ProcInfo(logger)

    print("first update")
    pi.update()
    print("Sleeping to accumulate")
    time.sleep(1)
    pi.update()

    print("System Monitoring:")
    sys_cpu_params = ['cpu_usr', 'cpu_sys', 'cpu_idle', 'cpu_nice', 'cpu_usage', 'context_switches', 'interrupts']
    sys_io_params = ['blocks_in', 'blocks_out', 'swap_in', 'swap_out']
    sys_mem_params = ['mem_used', 'mem_free', 'total_mem', 'mem_usage']
    sys_swap_params = ['swap_used', 'swap_free', 'total_swap', 'swap_usage']
    sys_load_params = ['load1', 'load5', 'load15', 'processes', 'uptime']
    sys_gen_params = ['hostname', 'cpu_MHz', 'no_CPUs', 'cpu_vendor_id', 'cpu_family', 'cpu_model', 'cpu_model_name', 'bogomips']
    sys_net_params = ['net_in', 'net_out', 'net_errs', 'ip']
    sys_net_stat = ['sockets_tcp', 'sockets_udp', 'sockets_unix', 'sockets_icm']
    sys_tcp_details = ['sockets_tcp_ESTABLISHED', 'sockets_tcp_SYN_SENT', 'sockets_tcp_SYN_RECV', 'sockets_tcp_FIN_WAIT1', 'sockets_tcp_FIN_WAIT2', 'sockets_tcp_TIME_WAIT', 'sockets_tcp_CLOSED', 'sockets_tcp_CLOSE_WAIT', 'sockets_tcp_LAST_ACK', 'sockets_tcp_LISTEN', 'sockets_tcp_CLOSING', 'sockets_tcp_UNKNOWN']

    print("sys_cpu_params", pi.getSystemData(sys_cpu_params))
    print("sys_io_params", pi.getSystemData(sys_io_params))
    print("sys_mem_params", pi.getSystemData(sys_mem_params))
    print("sys_swap_params", pi.getSystemData(sys_swap_params))
    print("sys_load_params", pi.getSystemData(sys_load_params))
    print("sys_gen_params", pi.getSystemData(sys_gen_params))
    print("sys_net_params", pi.getSystemData(sys_net_params))
    print("sys_net_stat", pi.getSystemData(sys_net_stat))
    print("sys_tcp_details", pi.getSystemData(sys_tcp_details))

    job_pid = os.getpid()

    print("Job (mysefl) monitoring:")
    pi.addJobToMonitor(job_pid, os.getcwd())
    print("Sleep another second")
    time.sleep(1)
    pi.update()

    job_cpu_params = ['run_time', 'cpu_time', 'cpu_usage']
    job_mem_params = ['mem_usage', 'rss', 'virtualmem', 'open_files']
    job_disk_params = ['workdir_size', 'disk_used', 'disk_free', 'disk_total', 'disk_usage']
    time.sleep(10)
    print("job_cpu_params", pi.getJobData(job_pid, job_cpu_params))
    print("job_mem_params", pi.getJobData(job_pid, job_mem_params))
    print("job_disk_params", pi.getJobData(job_pid, job_disk_params))

    pi.removeJobToMonitor(os.getpid())
